#ifndef __chips_ZR36016_H
#define __chips_ZR36016_H
/*
 * Copyright (c) 1997 Picture Elements, Inc.
 *    Stephen Williams (steve@picturel.com)
 *
 *    This source code is free software; you can redistribute it
 *    and/or modify it in source code form under the terms of the GNU
 *    General Public License as published by the Free Software
 *    Foundation; either version 2 of the License, or (at your option)
 *    any later version. In order to redistribute the software in
 *    binary form, you will need a Picture Elements Binary Software
 *    License.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *  ---
 *    You should also have recieved a copy of the Picture Elements
 *    Binary Software License offer along with the source. This offer
 *    allows you to obtain the right to redistribute the software in
 *    binary (compiled) form. If you have not received it, contact
 *    Picture Elements, Inc., 777 Panoramic Way, Berkeley, CA 94704.
 */
#ident "$Id: ZR36016.h,v 1.1 1998/01/19 02:34:17 steve Exp $"

/*
 * This class is a driver for the Zoran ZR36016 Integrated Color Space/-
 * Raster-to-Block Converter chip. For detailed description of the
 * programming interface for the chip in general, see the ZR36016 data
 * sheet available from ZORAN Compression Solutions <http://www.zoran.com>.
 *
 * The chip has two direct address registers, and 12 indirect
 * registers. The direct-address registers are implicitly manipulated
 * through the feature methods. All the indirect registers of the part
 * are manipulated transparently by the *_register methods.
 */

class ZR36016 {

    public:
	// Construct a ZR36016 driver. The base is the address of the
	// first byte of its address space, and the stride is the
	// number of bytes needed to get to the next register.
      explicit ZR36016(volatile unsigned char*base, unsigned stride=1)
      : base_(base), stride_(stride)
      { }

    public:
	// ZR36016 has a 4-bit virsion number.
      unsigned get_chip_version() const { return (base_[0] >> 4) & 0xf; }


	// Start an operation going, or test if an operation is in
	// progress. If the chip is set in single-frame mode, then the
	// go_flag() will be turned false when the current image is
	// finished.

      void go_flag(bool flag) { base_[0] = flag? 1 : 0; }
      bool go_flag()          { return (base_[0] & 1) == 1; }


	// These members manipulate the MODE register of the
	// part. This byte has 3 fields, with possible values from the
	// passed parameters. See the datasheet for details.

      enum PXOUT { PXOUT_RGB=0x00, PXOUT_444=0x20, PXOUT_422=0x40 };

      void set_mode(bool exp_flag, PXOUT po, unsigned mode)
	    { base_[stride_] = po | mode | (exp_flag? 0 : 0x80); }


	// set_width and set_height control the image window area
	// registers. set_width sets the number of pixels/line, and
	// optionally the left margin. The set_height sets the height
	// of the image area, with an optional top margin. If the
	// height is set to 0, the part will use the VIN signal to
	// mark the vertical bounds of the image.

      void set_width(unsigned pixels, unsigned margin =0)
	    { base_[2*stride_] = 2;
	      base_[3*stride_] = margin&0xff;
	      base_[3*stride_] = margin>>8;
	      base_[3*stride_] = pixels&0xff;
	      base_[3*stride_] = pixels>>8;
	    }

      void set_height(unsigned pixels, unsigned margin =0)
	    { base_[2*stride_] = 6;
	      base_[3*stride_] = margin&0xff;
	      base_[3*stride_] = margin>>8;
	      base_[3*stride_] = pixels&0xff;
	      base_[3*stride_] = pixels>>8;
	    }

      unsigned get_lines() const
	    { base_[2*stride_] = 10;
	      unsigned low = base_[3*stride_];
	      return (base_[3*stride_] << 8) + low;
	    }

    public:
	// These allow access to the INDIRECT registers of the
	// device. The contents of each register is an unsigned char
	// value, with an address from 0 - 11. The feature-based
	// methods above are preferred.

	// The load and save methods allow for efficient bulk saving
	// and loading of the register values.

      unsigned char get_register(unsigned addr) const
	    { base_[2*stride_] = addr; return base_[3*stride_]; }
      void set_register(unsigned addr, unsigned char val)
	    { base_[2*stride_] = addr; base_[3*stride_] = val; }

      void save_registers(unsigned char*buf) const
	    { base_[2*stride_] = 0;
	      for (unsigned idx = 0 ;  idx < 12 ;  idx += 1)
		    *buf++ = base_[3*stride_];
	    }

      void load_registers(const unsigned char*buf)
	    { base_[2*stride_] = 0;
	      for (unsigned idx = 0 ;  idx < 10 ;  idx += 1)
		    base_[3*stride_] = *buf++;
	    }

    private:
      volatile unsigned char*const base_;
      const unsigned stride_;

    private: // not implemented
      ZR36016(const ZR36016&);
      ZR36016& operator= (const ZR36016&);
};

/*
 * $Log: ZR36016.h,v $
 * Revision 1.1  1998/01/19 02:34:17  steve
 *  Add new chip drivers
 *
 */
#endif
