#ifndef __mc68322_ParallelPort_H
#define __mc68322_ParallelPort_H

/*
 * Copyright (c) 1995 Picture Elements Inc.
 *	Stephen Williams
 *
 *    This source code is free software; you can redistribute it
 *    and/or modify it in source code form under the terms of the GNU
 *    Library General Public License as published by the Free Software
 *    Foundation; either version 2 of the License, or (at your option)
 *    any later version. In order to redistribute the software in
 *    binary form, you will need a Picture Elements Binary Software
 *    License.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU Library General Public License for more details.
 *
 *    You should have received a copy of the GNU Library General Public
 *    License along with this program; if not, write to the Free
 *    Software Foundation, Inc.,
 *    59 Temple Place - Suite 330
 *    Boston, MA 02111-1307, USA
 *
 *    You should also have recieved a copy of the Picture Elements
 *    Binary Software License offer along with the source. This offer
 *    allows you to obtain the right to redistribute the software in
 *    binary (compiled) form. If you have not received it, contact
 *    Picture Elements, Inc.,
 *    777 Panoramic Way
 *    Berkeley, CA 94704.
 */
#ident "$Id: ParallelPort.h,v 1.11 1996/10/28 03:45:18 steve Exp $"
/*
 * $Log: ParallelPort.h,v $
 * Revision 1.11  1996/10/28 03:45:18  steve
 *  License notice.
 *
 * Revision 1.10  1995/11/02 07:30:21  steve
 *  More convenient interface to id string.
 *
 * Revision 1.9  1995/11/02 06:57:14  steve
 *  Split up the ParallelPort source,
 *  Support nibble mode id strings.
 *
 * Revision 1.8  1995/10/10 01:58:49  steve
 *  write takes pointer to constant.
 *
 * Revision 1.7  1995/10/08 04:27:52  steve
 *  Get rid of useless member.
 *
 * Revision 1.6  1995/10/04 23:17:15  steve
 *  Add RLE support to ECP mode.
 *
 * Revision 1.5  1995/10/04 07:17:01  steve
 *  Bum a little more performance out of reverse mode
 *  ECP, and potentially other reverse modes.
 *
 * Revision 1.4  1995/10/03 03:36:09  steve
 *  More 1284 more support.
 *
 * Revision 1.3  1995/09/25 01:38:45  steve
 *  Introduce IEE 1284 support, including nibble
 *  mode and byte mode.
 *
 * Revision 1.2  1995/09/08 23:18:25  steve
 *  change the ParallelPort to read ahead using PDMA.
 *  This should speed things up quite a bit, and allow
 *  the port to run independent of the CPU.
 *
 * Revision 1.1  1995/09/02 18:27:07  steve
 *  Initial import of UCR core to cvs.
 *
 *
 */

/*
 * This class drives the MC68322 parallel port interface, complete
 * with the PDMA interface. To construct, supply the interrupt level
 * and the 68322 register set base (if moved. I suggest sticking with
 * the default location, as that is one less mystery when things get
 * weird.)
 *
 * When creating a ParallelPort object, you must supply a buffer for
 * it to use as a read-ahead buffer. The ParallelPort reads ahead as
 * much as it can when the CPU is away doing other things, and it uses
 * the PDMA controller to do that magic. If you make the buffer
 * larger, it can read ahead more. But, due to alignment hazards, it
 * will stop reading ahead if you read some bytes out of
 * it. Therefore, too large a buffer and you will just waste
 * memory. Choose a size about equal to the largest single read and
 * you will do well.
 */

# include  <ucr/ISync.h>
# include  <stdlib.h>


class ParallelPort {

        // the buf is a buffer that the ParallelPort will use for
	// transfers. Note that the buffer must be in DRAM, or PDMA
	// will not work.
    public:
      ParallelPort(void*buf, size_t nbuf, unsigned char level,
		   unsigned long base =0xfff000);
      ~ParallelPort();


        // Read the specified number of bytes into the specified
	// buffer. The count is returned, and will only be less then
	// the requested amount if there was a transmission error. The
	// read operation will block until the data is ready.
      size_t read(void*buffer, size_t count);

        // Write the specified number of bytes from the buffer to the
	// host. The write will only block if it cannot write
	// everything into the outbound buffer. It will only return
	// less then count if there is an I/O error.
      size_t write(const void*buffer, size_t count);

        // Set the ID string to return to the host. The length is
	// required as non-ascii stuff may be here. The caller must
	// supply such things as the length code.
      void set_id_string(unsigned short count, const char*bytes);

    private:
      ISync in_sync_;
      bool  in_flag_; // True if the interface is set up for input.

        // Mark the boundries of the input buffer.
      unsigned char*const buf_;
      const size_t nbuf_;

        // These reflect how many bytes are held, and how many are
	// pending.
      size_t transfer_;
      size_t holding_;
      unsigned char*readp_;

        // This is false when the PDMA has control of the port.
      bool dma_ready_;

      void copyout(unsigned char*, size_t);
      void dma_go(size_t);


        // These members and methods manage the operation (what mode
	// am I in?) of the parallel port. With IEEE1284, this is a
	// non-trivial task.
    private:

      void terminate_ieee1284_();
      void enter_nibble_mode_();
      void enter_nibble_id_mode_();
      void enter_byte_mode_();
      void enter_ecp_mode_();

        // This method pointer allows the write to awaken any of the
	// various reverse idle phases that I might be in. This will
	// be set to 0 if the port is not in a state where that's a
	// useful thing to do.
      void (ParallelPort::*reverse_transfer_)();
      void reverse_nibble_();
      void reverse_byte_();
      void reverse_ecp_();

        // These members keep track of what is going on outbound so
	// that the negotiation can know how to set the nFault line,
	// and to awaken things if necessary.
      static bool test_for_outbound_(ParallelPort*);
      ISync out_sync_;
      size_t out_remaining_;
      const unsigned char*out_buffer_;

      void reverse_nibble(unsigned char);

        // These members track sending of the ID string.
      const char*id_;
      const char*curid_;
      unsigned short nid_;
      enum { LOW, HIGH, LOWS1, HIGHS1, LOWS2, HIGHS2 } nibble_sel_;

    private:
        // Handle non-dma interrupts depending on my mode.
      void (ParallelPort::*port_isr_)();
      void port_isr_compat_();
      void port_isr_error_();
      void port_isr_negotiate_();
      void port_isr_nibble_();
      void port_isr_nibble_id_();
      void port_isr_byte_();
      void port_isr_ecp_forward_();
      void port_isr_ecp_reverse_();
      bool rle_flag_;

      static void nap_for_host_();

    private:
      struct REGS_ {
	    volatile unsigned short ppcr;
	    volatile unsigned short lines;
	    volatile unsigned char  cmd_bit;
	    volatile unsigned char  data;
	    unsigned short ack_width;
      };

      struct INTR_ {
	    volatile unsigned short enable_flags;
	    volatile unsigned short intr_flags;
	    volatile unsigned short level;
      };

      struct PDMA_ {
	    volatile void*  transfer_address;
	    volatile unsigned short transfer_count;
	    volatile unsigned short flush_request;
      };

      // PDMA interrupt regs are arranged the same as INTR_.

      REGS_* const port_regs_;
      INTR_* const port_intr_;
      PDMA_* const pdma_regs_;
      INTR_* const pdma_intr_;


      static void isr_(ParallelPort*, unsigned);

    private: // Not implemented
      ParallelPort(const ParallelPort&);
      ParallelPort& operator= (const ParallelPort&);
};

#endif
