#ifndef __ucr_Ticker_H
#define __ucr_Ticker_H

/*
 * Copyright (c) 1995 Picture Elements Inc.
 *	Stephen Williams
 *
 *    This source code is free software; you can redistribute it
 *    and/or modify it in source code form under the terms of the GNU
 *    Library General Public License as published by the Free Software
 *    Foundation; either version 2 of the License, or (at your option)
 *    any later version. In order to redistribute the software in
 *    binary form, you will need a Picture Elements Binary Software
 *    License.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU Library General Public License for more details.
 *
 *    You should have received a copy of the GNU Library General Public
 *    License along with this program; if not, write to the Free
 *    Software Foundation, Inc.,
 *    59 Temple Place - Suite 330
 *    Boston, MA 02111-1307, USA
 *
 *    You should also have recieved a copy of the Picture Elements
 *    Binary Software License offer along with the source. This offer
 *    allows you to obtain the right to redistribute the software in
 *    binary (compiled) form. If you have not received it, contact
 *    Picture Elements, Inc.,
 *    777 Panoramic Way
 *    Berkeley, CA 94704.
 */
#ident "$Id: Ticker.h,v 1.6 1996/10/28 03:45:25 steve Exp $"
/*
 * $Log: Ticker.h,v $
 * Revision 1.6  1996/10/28 03:45:25  steve
 *  License notice.
 *
 * Revision 1.5  1995/10/20 04:02:40  steve
 *  Make timeouts lighter, less buggy, and a
 *  little less general. Light is better here.
 *  Also, leave lists as a general tool.
 *
 * Revision 1.4  1995/09/29 17:32:30  steve
 *  Define NS resolution of ticker.
 *
 * Revision 1.3  1995/09/25 00:28:27  steve
 *  Optimize ISR selection and dispatch, and try
 *  to speed up handling of time.
 *
 * Revision 1.2  1995/09/17 07:00:29  steve
 *  Add support for timeouts.
 *
 * Revision 1.1  1995/09/02 18:27:11  steve
 *  Initial import of UCR core to cvs.
 *
 *
 */


/*
 * The ticker represents the typical embedded timepiece, a real time
 * clock that ticks at regular intervals and calls handlers to do
 * various jobs during time ticks. The programmer can set handlers
 * that tick at regular intervals, or can set a one time timeout
 * event. Either way, it is the derived class that fixes the rate that
 * the clock ticks.
 *
 * It is up to the application to create exactly one concrete instance
 * that matches the actual device. That instance sets the resolution
 * of the clock.
 *
 * Where needed, methods of this class will use thread_splhi to
 * protect themselves from the handle_tick() method. This naturally
 * means that the timer tick must be on a maskable interrupt. (Or no
 * interrupt at all.)
 *
 * Try not to make excessive use of the timeout method, as each tick
 * needs to go through the entire list decrementing counters, and a
 * long list can cause performance trouble.
 */



class Ticker {

    public:
      Ticker(unsigned long ns_per_tick);
      virtual ~Ticker() =0;
      virtual void start() =0;

        // Get the resolution of this ticker.
      unsigned long ns_per_tick() const { return ns_per_tick_; }


      void set_tick_handler( void(*fun)(void*), void*arg );

        // Call the function after some number of ticks. How long a
	// tick is depends on the interval between ticks, set by the
	// derived class. Take the number of ns you want, and divide
	// it by ns_per_tick() to get the desired number of ticks.
      void timeout(unsigned ticks, void(*fun)(void*), void*arg);

        // This form simply blocks until the timeout expires. It uses
	// the above method and is just a convenience.
      void timeout(unsigned ticks);


        // The specific device that implements the ticker calls this
	// method to handle ticks. This method may be called in the
	// isr for the clock interrupt.
    protected:
      void handle_tick();

    private:
      typedef void (*fun_t)(void*);
      enum { TICKER_FUN_LIMIT = 4, CALLOUT_LIMIT = 4 };
      fun_t fun_[TICKER_FUN_LIMIT];
      void* arg_[TICKER_FUN_LIMIT];

      struct TimeoutCell {
	    unsigned ticks;
	    void(*fun)(void*);
	    void*arg;
      };

      TimeoutCell callout_[CALLOUT_LIMIT];
      unsigned ncallout_;

      const unsigned long ns_per_tick_;

    private: // not supported
      Ticker(const Ticker&);
      Ticker& operator= (const Ticker&);
};

#endif
