/*
 * Copyright (c) 1995 Picture Elements Inc.
 *	Stephen Williams
 *
 *    This source code is free software; you can redistribute it
 *    and/or modify it in source code form under the terms of the GNU
 *    Library General Public License as published by the Free Software
 *    Foundation; either version 2 of the License, or (at your option)
 *    any later version. In order to redistribute the software in
 *    binary form, you will need a Picture Elements Binary Software
 *    License.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU Library General Public License for more details.
 *
 *    You should have received a copy of the GNU Library General Public
 *    License along with this program; if not, write to the Free
 *    Software Foundation, Inc.,
 *    59 Temple Place - Suite 330
 *    Boston, MA 02111-1307, USA
 *
 *    You should also have recieved a copy of the Picture Elements
 *    Binary Software License offer along with the source. This offer
 *    allows you to obtain the right to redistribute the software in
 *    binary (compiled) form. If you have not received it, contact
 *    Picture Elements, Inc.,
 *    777 Panoramic Way
 *    Berkeley, CA 94704.
 */
#ident "$Id: PrintEngine.cc,v 1.9 1996/10/28 03:45:56 steve Exp $"
/*
 * $Log: PrintEngine.cc,v $
 * Revision 1.9  1996/10/28 03:45:56  steve
 *  License notice.
 *
 * Revision 1.8  1996/02/22 22:25:04  steve
 *  Enable the PBB interrupt
 *
 * Revision 1.7  1995/12/14 22:19:09  steve
 *  Provide page synchronization.
 *
 * Revision 1.6  1995/11/07 07:54:49  steve
 *  Support optional smart-toner mode.
 *
 * Revision 1.5  1995/10/25 04:04:49  steve
 *  Support Automatic printer resolution switching.
 *
 * Revision 1.4  1995/10/17 18:29:04  steve
 *  Put in the band queue class help for
 *  doing queue synchronization.
 *
 * Revision 1.3  1995/10/09 01:28:58  steve
 *  Keep slightly better track of errors.
 *
 * Revision 1.2  1995/09/11 18:43:05  steve
 *  Cure extra page at end of job.
 *
 * Revision 1.1  1995/09/02 18:27:19  steve
 *  Initial import of UCR core to cvs.
 *
 *
 */


# include  <mc68322/PrintEngine.h>
# include  <ucr/ucr.h>
# include  <assert.h>

/*
 * In some of the following comments, "Bandit Manual" refers to the
 * "MC68322 Integrated Printer Processor User's Manual."
 *
 * This is in the mc68322 library because I'm hoping to get all the
 * engine specific stuff out of here. It turns out that only the video
 * geometry is engine specific.
 */


const unsigned short EVENT_BAD_ADDRESS= 0x0001;
const unsigned short EVENT_VIDEO_UNDER= 0x0002;
const unsigned short EVENT_BAND_UNDER = 0x0004;
const unsigned short EVENT_PAGE_DONE  = 0x0008;
const unsigned short EVENT_BAND_BEGIN = 0x0010;
const unsigned short EVENT_BAND_FULL  = 0x0020;
const unsigned short EVENT_BAND_BUSY  = 0x0040;


PrintEngine::PrintEngine(const Geometry*speeds, unsigned char nspeeds,
			 unsigned char level, unsigned long base)
: pvc_ ( (PVC_*)  (base + 0x400) ),
  pcom_( (PCOM_*) (base + 0x500) ),
  intr_( (INTR_*) (base + 0x700) ),
  speeds_(speeds), nspeeds_(nspeeds), ctrl_mask_(0x1401),
  com_status_(0), com_status_ready_(true), top_of_page_(true),
  printing_begun_(0)
{
        // Soft reset of the PVC interface.
        // See 5.5 (p. 5-15) of the Bandit Manual
      *(unsigned short*)(base + 0x7a0) = 1;

        // Use the internal PLL to generate a 30Mhz/8 video dot clock.
        // See 10.1.2 (p 10-3) of the Bandit Manual.
      pvc_->pvc_flags = speeds_[0].ctrl | ctrl_mask_;

        // See 10.1.3 (p. 105) of the Bandit Manual.
      pvc_->vmargin = speeds_[0].top_margin;
      pvc_->hmargin = speeds_[0].left_margin;

        // The MC68322 supplies command and status clocks, and uses
	// seperate data lines for each. Set the clock divisor to slow
	// the strobing a little bit. These values are not optional:
	// they are fixed by the hardware choice.

        // See 10.1.1 (p. 10-2) of the Bandit Manual
      pcom_->pcom_flags = 0x000b;


        // Disable interrupts, set the level, and clear any pending
	// interrupts, which shouldn-t exist anyhow.

        // See 10.1.4 (p. 10.6) of the Bandit Manual
      intr_->event_level = level;
      intr_->event_flags = 0x001f;

      intr_->com_level  = level;
      intr_->com_flags  = 0x0003;

      set_isr(level, intr, this);

        // Now turn on the interrupts that apply to this device.

      intr_->com_enable   = 1;
      intr_->event_enable = EVENT_PAGE_DONE | 0x1f;

        // Send a command to the printer forcing it to accept the
	// controller (me!) as the source of the status strobe.
      do_command(0x40, false);
}


unsigned char PrintEngine::do_command(unsigned char cmd, bool wait_flag)
{
        // Clear the status from any previous command.
      get_status();

      com_status_ready_ = false;
      pcom_ -> command  = cmd;

      if (wait_flag) return get_status();
      else return 0;
}

/*
 * Return the most recent status word from the printer. If there is a
 * command in progress, wait for it.
 */
unsigned char PrintEngine::get_status() const
{
      if (!com_status_ready_) com_sync_.sleep(&com_status_ready_);
      assert( com_status_ready_ );
      return com_status_;
}

void PrintEngine::intr_com_(unsigned short com_flags)
{
        // A command completed and the status has come back. Turn off
	// the talking flag and awaken blocked threads.
      if (com_flags & 1) {
	    assert( ! com_status_ready_ );
	    com_status_ = pcom_->status & 0xff;
	    com_status_ready_ = true;
	    com_sync_.wakeup();
      }
}


void PrintEngine::print_band(Band*band)
{
      unsigned short cookie = thread_splhi();
      pending_queue_.enqueue(band);
      load_bands();
      thread_spl(cookie);
}


Band* PrintEngine::pull_completed_band()
{
      print_sync_.sleep(&BandQueue::not_empty, &done_queue_);

      unsigned short cookie = thread_splhi();
      Band* result = done_queue_.pull();
      thread_spl(cookie);

      return result;
}

void PrintEngine::page_sync()
{
      psync_flag_ = false;
      psync_sync_.sleep(&psync_flag_);
}

void PrintEngine::intr_video_(unsigned short event_flags)
{
      assert( ! (event_flags & EVENT_BAD_ADDRESS) );
      assert( ! (event_flags & EVENT_VIDEO_UNDER) );
      assert( ! (event_flags & EVENT_BAND_UNDER)  );

        // A BAND_BEGIN means that a new band has started. This means
	// that a previous band is finished. Keep track of the begun
	// bands with the printing_begun_ member.

        // A PAGE_DONE means that any begun band is finished, but not
	// that any new bands were begun.

      if (event_flags & (EVENT_BAND_BEGIN|EVENT_PAGE_DONE)) {
	    bool wakeup_flag = false;

	    while (printing_begun_ > 0) {
		  printing_begun_ -= 1;
		  Band*bp = printing_queue_.pull();
		  done_queue_.enqueue(bp);
		  wakeup_flag = true;
	    }

	    if (event_flags & EVENT_BAND_BEGIN) {
		  printing_begun_ += 1;
		  psync_flag_ = true;
		  psync_sync_.wakeup();
	    }

	    load_bands();

	    if (wakeup_flag) print_sync_.wakeup();
      }
}


/*
 * This does most of the band manipulation work of the video
 * interface. It is called every time there is a chance that I might
 * start more bands going, which I do by loading registers with the
 * band information and putting it in the printing_queue_. I also
 * make sure the printer is going as long as there are bands to print.
 *
 * This method also tracks the top-of-page and supplies a margin as
 * needed, so that the application never has to worry about margins.
 */
void PrintEngine::load_bands()
{
      Band*bp = 0;

      while ((! (intr_->event_flags & EVENT_BAND_FULL)) &&
	     (bp = pending_queue_.pull())) {

	    const Geometry&geo = speeds_[bp->geometry()];
	    printing_queue_.enqueue(bp);

	      // Load the geometry for this band, including horizontal
	      // resolution and margins which come from the spec table.
	    pvc_->pvc_flags = geo.ctrl | ctrl_mask_;
	    pvc_->vmargin = top_of_page_ ? geo.top_margin : 0;
	    pvc_->hmargin = geo.left_margin;
	    pvc_->width   = bp->width();
	    pvc_->height  = bp->height();

	    pvc_->pcb_flags = (bp->end_of_page()? 0 : 1)
	                     |(bp->smart_toner()? 4 : 0);

	    pvc_->image_bit_address = ((unsigned long)bp->bits()) << 3;

	    top_of_page_ = bp->end_of_page();
      }

      if ((printing_queue_.count()-printing_begun_) > 0) pvc_->print = 1;
      else pvc_->print = 0;
}


/*
 * Handle all the interrupts from this device through on interrupt
 * handler. Note that this device may share the interrupt number with
 * other devices so be careful and make no assumptions.
 */
void PrintEngine::intr(PrintEngine*obj, unsigned)
{
      unsigned short com_flags = obj->intr_->com_flags;
      unsigned short event_flags = obj->intr_->event_flags;

      obj->intr_->event_flags = event_flags & 0x1f;
      obj->intr_->com_flags = com_flags;

      if (event_flags) obj->intr_video_(event_flags);
      if (com_flags) obj->intr_com_(com_flags);
}

