
/*
 * Copyright (c) 1995 Picture Elements Inc.
 *	Stephen Williams
 *
 *    This source code is free software; you can redistribute it
 *    and/or modify it in source code form under the terms of the GNU
 *    Library General Public License as published by the Free Software
 *    Foundation; either version 2 of the License, or (at your option)
 *    any later version. In order to redistribute the software in
 *    binary form, you will need a Picture Elements Binary Software
 *    License.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU Library General Public License for more details.
 *
 *    You should have received a copy of the GNU Library General Public
 *    License along with this program; if not, write to the Free
 *    Software Foundation, Inc.,
 *    59 Temple Place - Suite 330
 *    Boston, MA 02111-1307, USA
 *
 *    You should also have recieved a copy of the Picture Elements
 *    Binary Software License offer along with the source. This offer
 *    allows you to obtain the right to redistribute the software in
 *    binary (compiled) form. If you have not received it, contact
 *    Picture Elements, Inc.,
 *    777 Panoramic Way
 *    Berkeley, CA 94704.
 */
#ident "$Id: heap.cc,v 1.8 1998/04/11 23:42:41 steve Exp $"

# include  <ucr/heap.h>

/*
 * Keep a sorted list of cells in memeory. The cells can be allocated
 * or free, but they live in a contiguous segment. The free chunks of
 * memory are kept in a sorted circular doubly linked list, with cells
 * merged whenever possible. The root cell can never be allocated or
 * merged.
 *
 * The heap handler code has the special property that it tries to
 * keep allocations as low in memory as practical, and does not touch
 * memory beyond what has been allocated. This allows these heap
 * routines to be used to manage a heap that shares space with a
 * processor stack that grows down from the end of memory.
 *
 *    A NOTE ON UNITS: The HeapCell structure is 8 bytes, and is also
 *    the unit of size. If a cell is 64 bytes (including the header)
 *    its size is given as 8.
 *
 * If the cell is free, next points to the next free cell and the
 * value cell[1].next points to the previous free cell. Allocating
 * cells keeps this list up to date.
 *
 * If the cell is allocated, cell->next points to the cell directly
 * after. The deallocator can follow ->next until it finds the next
 * free cell, which will be the first free cell after it in the heap.
 *
 * The heap looks like this when first initialized:
 *
 * root--+            +-------------------+
 *       |            |                   |
 *       v            v                   |
 *       +--------+---------------+       |
 *       |        | size=2 free=2 | ---+  |
 *       +--------+---------------+    |  |
 *   +-- |        | xxxxxxxxxxxxx |    |  |
 *   |   +========+===============+ <--+  |
 *   +-->|        | size=N free=1 | ------+
 *       +--------+---------------+
 *       |        | xxxxxxxxxxxxx |
 *       +--------+---------------+
 *
 * The root cell always has a size of 2 and is marked with free=2. The
 * root cell is special and always sits in the beginning of the
 * heap. The other cell is the only free cell and contains all the
 * free space of the heap. (Minus the 2 of the root, of course.) The
 * next pointer of the root cell always points to the first free cell,
 * and the prev pointer (root[1].next) always points to the last free
 * cell.
 *
 * A free cell (and *only* a free cell) may be split in preparation
 * for allocation. If size=N (N*8 bytes including the header) is split
 * out of this cell:
 *
 *                                                     A
 *                    +--------+-----------------+     ^
 *              +---- |        | size=X free=1   |     |
 *              |     +--------+-----------------+     |
 *              |     |        | xxxxxxxxxxxxxxx | ----+
 *              v     +--------+-----------------+
 *              B
 *
 * The result is the two following cells. The next pointer of the cell
 * points to a new cell further on, that is linked in after the
 * original cell. The original cell contains the space that was split
 * out.
 *
 *                                                     A
 *                    +--------+-----------------+     ^
 *              +---- |        | size=N free=1   |     |
 *              |     +--------+-----------------+     |
 *              |     |        | xxxxxxxxxxxxxxx | ----+
 *              |     +--------+-----------------+ <---+
 *              |                                      |
 *              |                                      |
 *              |           ...N-2 of space...         |
 *              |                                      |
 *              |                                      |
 *              +---> +--------+-----------------+     |
 *              +---- |        | size=X-N free=1 |     |
 *              |     +--------+-----------------+     |
 *              |     |        | xxxxxxxxxxxxxxx | ----+
 *              v     +--------+-----------------+
 *              B
 *
 *
 * Note that by this technique, the size field, when added to the base
 * address of the cell, always gets the pointer to the cell after.
 *
 * Within the free list, the next and prev pointers always point to
 * free cells.
 *
 * Allocation works by taking the desired cell, splitting it to the
 * right size (using the above technique) and unlinking it from the
 * list. The value returned is the pointer to the second half of the
 * cell structure, because only the first 2 words are needed at this
 * point. The next pointer is adjusted to point to the end (i.e. the
 * next cell) of the cell. The size and pointer together form a good
 * position-dependent magic number for the cell.
 *
 * Once allocation starts happening, the free list of cells will skip
 * over allocated regions, so the next pointers and size will not
 * necessarily correspond. However, the prev pointer of the next free
 * cell will point to the current free cell, so the redundancy is
 * available there.
 *
 * Deallocation attempts to reintroduce the cell to the free list, and
 * merge cells if they are adjacent. When the deallocator gets a
 * pointer to an allocated cell, it searches the heap space for the
 * next free cell above it. It uses its next pointer to get
 * started. It continues following next pointers until it gets to a
 * cell that is free.
 *
 * If the cell's next pointer points to a free cell right away, then
 * happy day, this cell can be joined to its following cell immediately.
 */

# include  "__heap.h"
# include  <assert.h>

/*
 * Initialize the heap for use by the memory allocator. Pass a pointer
 * to the space to use as a heap, and the size in bytes.
 */
extern "C" HEAP_SPACE* heap_init(void*space, unsigned long size)
{
      size = size / sizeof(HeapCell);
      HeapCell *root = (HeapCell*)space;
      root[0].next = root+2;
      root[0].free = 2;
      root[0].size = 2;
      root[1].next = root+2;

      HeapCell*cell = root+2;
      cell[0].next = root;
      cell[0].size = size - 2;
      cell[0].free = 1;
      cell[1].next = root;

      return (HEAP_SPACE*)space;
}

/*
 * Search for the first cell that fits. Split the cell if that makes
 * sense, and return a pointer to the memory part. Internally, these
 * routines manipulate memory in units of HeapCell size. These are
 * 64bit aligned, for your dereferencing pleasure.
 *
 * NOTE: In general purpose allocators, there is no evidence that
 * first fit is better or worse then best/worst fit so it is not worth
 * the effort of implementing the latter. However, keeping memory
 * sorted and merged is a big win for fast garbage collection.
 */
void* heap_alloc(unsigned long size, HEAP_SPACE*heap)
{
      HeapCell*root = (HeapCell*) heap;

        // Normalize the requested size.

      unsigned long hsize = (size + 2*sizeof(HeapCell) - 1) /
	sizeof(HeapCell);

        // First fit search.
      HeapCell*cur = root->next;
      while ( ((unsigned long)cur->size < hsize) && (cur != root) )
	cur = cur->next;

      if ( cur == root ) return 0;


        // Trim excess off the located cell.
      if ((unsigned long)cur->size > (2+hsize)) {

	    HeapCell*split = cur+hsize;
	    split[0].next = cur->next;
	    split[0].next[1].next = split;
	    split[0].size = cur->size - hsize;
	    split[0].free = 1;
	    split[1].next = cur;

	    cur->next = split;
	    cur->size = hsize;
      }

	// There MUST be a free cell after me. This acts like a
	// trailer for other parts of the allocator.
      if (cur->next == root) return 0;

        // Pull this cell out of the free list.

      cur[1].next->next = cur->next;
      cur->next[1].next = cur[1].next;
      cur->free = 0;

	// Point to the cell after me. It is possible for a cell
	// between me and my current free cell neighbor to be
	// released, so when I'm released I will need to search
	// starting right after me for a next cell.
      cur->next = cur + cur->size;

      return cur + 1;
}

/*
 * This routine assumes that there is a free cell before the end.
 */
void heap_free(void*memory)
{
      HeapCell*cell = ((HeapCell*)memory) - 1;

      HeapCell*cur = cell->next;

      assert( cell->free == 0 );
      cell->free  = 1;

        // Link to the free cell after this. If the immediate
	// following cell is free, then merge immediately. Otherwise,
	// search for the next free cell and do not merge. Also, do
	// not merge with the root cell.
      if (cur->free == 1) {
	    cell->size += cur->size;
	    cell->next  = cur->next;
	    cell[1].next= cur[1].next;
      } else {
	    while (! cur->free) cur = cur->next;
	    assert( cur->free == 1 );
	    cell->next = cur;
	    cell[1].next = cur[1].next;
      }

      cell[1].next->next = cell;
      cell->next[1].next = cell;

        // Attempt to join with the previous cell.
      cur = cell[1].next;
      if ( (cur->free == 1) && (cur + cur->size == cell)) {
	    cur->size += cell->size;
	    cur->next  = cell->next;
	    cur->next[1].next = cur;
      }

}

/*
 * $Log: heap.cc,v $
 * Revision 1.8  1998/04/11 23:42:41  steve
 *  Fix allocation error.
 *
 * Revision 1.7  1998/04/09 23:40:39  steve
 *  Some inexpensive error checking.
 *
 * Revision 1.6  1996/11/11 03:48:04  steve
 *  Optimize for size.
 *
 * Revision 1.5  1996/10/28 03:46:24  steve
 *  License notice.
 *
 * Revision 1.4  1995/10/15 23:59:19  steve
 *  Fix free cell splitting in the allocate,
 *  Add free space statistic,
 *  Make new() operators thread safe.
 *
 * Revision 1.3  1995/09/08 05:58:21  steve
 *  Better integration of heap handling with
 *  the new/delete operators.
 *
 * Revision 1.2  1995/09/07 18:25:10  steve
 *  Seperate heap_selftest from other routines.
 *
 * Revision 1.1  1995/09/02 18:27:29  steve
 *  Initial import of UCR core to cvs.
 */
