/*
 * Copyright (c) 1996 Picture Elements, Inc.
 *    Stephen Williams
 *
 *    This source code is free software; you can redistribute it
 *    and/or modify it in source code form under the terms of the GNU
 *    Library General Public License as published by the Free Software
 *    Foundation; either version 2 of the License, or (at your option)
 *    any later version. In order to redistribute the software in
 *    binary form, you will need a Picture Elements Binary Software
 *    License.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU Library General Public License for more details.
 *
 *    You should have received a copy of the GNU Library General Public
 *    License along with this program; if not, write to the Free
 *    Software Foundation, Inc.,
 *    59 Temple Place - Suite 330
 *    Boston, MA 02111-1307, USA
 *
 *    You should also have recieved a copy of the Picture Elements
 *    Binary Software License offer along with the source. This offer
 *    allows you to obtain the right to redistribute the software in
 *    binary (compiled) form. If you have not received it, contact
 *    Picture Elements, Inc.,
 *    777 Panoramic Way
 *    Berkeley, CA 94704.
 */
#ident "$Id: thread_regs-i960.cc,v 1.2 1996/10/28 03:46:42 steve Exp $"
/*
 * $Log: thread_regs-i960.cc,v $
 * Revision 1.2  1996/10/28 03:46:42  steve
 *  License notice.
 *
 * Revision 1.1  1996/09/19 04:44:17  steve
 *  Add the i960_ise target, change the target
 *  configuration headers to allow sharing by
 *  targets, fix up the make clean rule.
 *
 * Revision 1.7  1996/08/24 20:57:36  steve
 *  The priority cookie is not necessarily short, so make
 *  a typedef for it.
 *
 *  There is no need (indeed no way) to save and restore the
 *  fp of a thread, so put the PFP, which is really the fp,
 *  into the fp register slot.
 *
 *  Slightly better check of the stack frame at thread switch
 *  time. Also have faults check for some bad things.
 *
 * Revision 1.6  1996/08/22 00:34:45  steve
 *  Fix retrieval of local registers
 *
 * Revision 1.5  1996/08/18 22:36:25  steve
 *  Switch hte TC register with the thread.
 *
 * Revision 1.4  1996/08/15 02:58:58  steve
 *  Add thread_set_reg.
 *
 * Revision 1.3  1996/08/14 05:09:57  steve
 *  Get registers from fault_record if available.
 *
 * Revision 1.2  1996/08/06 22:22:37  steve
 *  Get the registers right.
 *
 * Revision 1.1  1996/08/03 03:30:46  steve
 *  Create the ucr/asm.h header as a target specific
 *  header with line functions, etc; Replace the
 *  __isr_run_flag with a bit of inline assembly code.
 *
 *  Move much thread code out of the __Secheduler class
 *  and into more public thread_ functions, to eliminate
 *  some wasted function call depth. Even the main_thread
 *  is move into its own file.
 *
 *  Keep a list of available threads for debug use, and
 *  add a function to get thread registers.
 *
 */

# include  <ucr/ucr.h>
# include  "__thread.h"
# include  <ucr/fault.h>


/*
 * For the i960, Write the registers into the buffer as follows:
 *	r0-r15,
 *	g0-g15,
 *	pcw, ac, tc, ip
 * This matches the order that GDB deals with them internally, so does
 * not burden the interface code with reordering things.
 */


/*
 * If the thread is faulty (trapped the FAULT_RECORD pointer is
 * present) then I try to get the frame information starting where the
 * fault frame is. This is a little tricky, but can be done.
 */
static void thread_fault_regs(THREAD*tid, unsigned long*buf)
{
	// On the i960, the FAULT_RECORD is just below the stack frame
	// where the frame for the handler goes. Thus, I start with
	// the pointer to the fault record, and get the fault frame
	// pointer by simple arithmetic.
      const FAULT_RECORD*fid = get_fault(tid);
      const unsigned long*ffp = (const unsigned long*)(fid+1);

	// The frame pointer of the faulted code can then be guessed
	// from the contents of the PFP register (r0) of the fault
	// handler.
      const unsigned long*fp = (const unsigned long*)(ffp[0] & ~0xf);

	// With the frame pointer, I can get the local registers of
	// the faulted frame.
      for (unsigned idx = 0 ;  idx < 16 ;  idx += 1)
	    buf[idx] = fp[idx];

	// This next part is tricky. The uCR_fault_common function
	// puts the global registers of the faulted frame on the
	// stack, which starts right after the register save area for
	// the fault handler. 
      for (unsigned idx = 0 ;  idx < 16 ;  idx += 1)
	    buf[idx+16] = ffp[idx+16];

	// These remaining things come right out of the FAULT_RECORD.
      buf[31] = (unsigned long)fp;
      buf[32] = fid->pc;
      buf[33] = fid->ac;
      buf[34] = tid->cookie.tc;
      buf[35] = fp[2];
}

static void thread_nofault_regs(THREAD*tid, unsigned long*buf)
{

	// Get the local registers out of the thread stack. The pfp
	// holds a pointer to the frame where the registers live.

      unsigned long *fp = (unsigned long*)(tid->cookie.regs[15] & ~0xf);
      for (unsigned idx = 0 ;  idx < 16 ;  idx += 1)
	    buf[idx] = fp[idx];

	// Pull the global registers out of the thread register save
	// area, then retrieve the fp from the cookie.pfp member. (When
	// the thr_switch_ was called, g15 was replaced with a new fp
	// that contains at this point nothing of value. We want the
	// frame where thr_switch_ was called.) Also get the other
	// special registers.

      for (unsigned idx = 0 ;  idx < 15 ;  idx += 1)
	    buf[16+idx] = tid->cookie.regs[idx];

      buf[31] = (unsigned long)fp;
      buf[32] = tid->cookie.pc;
      buf[33] = tid->cookie.ac;
      buf[34] = tid->cookie.tc;
      buf[35] = fp[2];
}

void thread_regs(THREAD*tid, void*buffer)
{
      unsigned long*buf = (unsigned long*)buffer;

      if (get_fault(tid)) thread_fault_regs(tid, buf);
      else thread_nofault_regs(tid, buf);
}

static void thread_fault_set_reg(THREAD*tid, unsigned regno, unsigned long val)
{ 
      FAULT_RECORD*fid = tid->fault_;
      unsigned long*ffp = (unsigned long*)(fid+1);
      unsigned long*fp = (unsigned long*)(ffp[0] & ~0xf);
      unsigned long*pfp = (unsigned long*)(fp[0] & ~0xf);

      if (regno < 16) {
	    pfp[regno] = val;
	    return;
      }

      if (regno < 32) {
	    ffp[regno] = val;
	    return;
      }

      switch (regno) {
	  case 32:
	    fid->pc = val;
	    break;
	  case 33:
	    fid->ac = val;
	    break;
	  case 34:
	    tid->cookie.tc = val;
	    break;
	  case 35:
	    fp[2] = val;
	    break;
      }
}

static void thread_nofault_set_reg(THREAD*tid, unsigned regno,
				   unsigned long val)
{
}

void thread_set_reg(THREAD*tid, unsigned regno, unsigned long val)
{
      if (get_fault(tid)) thread_fault_set_reg(tid, regno, val);
      else thread_nofault_set_reg(tid, regno, val);
}
