/*
 * Copyright (c) 1997 Picture Elements, Inc.
 *    Stephen Williams (steve@picturel.com)
 *
 *    This source code is free software; you can redistribute it
 *    and/or modify it in source code form under the terms of the GNU
 *    General Public License as published by the Free Software
 *    Foundation; either version 2 of the License, or (at your option)
 *    any later version. In order to redistribute the software in
 *    binary form, you will need a Picture Elements Binary Software
 *    License.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *  ---
 *    You should also have recieved a copy of the Picture Elements
 *    Binary Software License offer along with the source. This offer
 *    allows you to obtain the right to redistribute the software in
 *    binary (compiled) form. If you have not received it, contact
 *    Picture Elements, Inc., 777 Panoramic Way, Berkeley, CA 94704.
 */
#ident "$Id: linux.c,v 1.4 1998/06/06 05:40:03 steve Exp $"

# include  "ucrif.h"
# include  "os.h"
# include  "t-config.h"
# include  "ucrpriv.h"
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0)
# include  <linux/poll.h>
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,1,0)
# define ioremap vremap
# define iounmap vfree
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0)
MODULE_PARM(ucr_major,"i");
MODULE_PARM_DESC(ucr_major,"Use this major number");

MODULE_PARM(debug_flag,"i");
MODULE_PARM_DESC(debug_flag,"Enable debug messages");
#endif

int atomic_sleep_on(struct Instance*xsp,
		     unsigned long mask, struct wait_queue**sync)
{
      unsigned long flags;
      save_flags(flags);
      cli();
      dev_unmask_irqs(xsp->dev, mask);
      sleep_on(sync);
      dev_mask_irqs(xsp->dev);
      restore_flags(flags);
      return 0;
}

void set_time_delay(struct Instance*xsp,
		    struct timer_list*tim,
		    void (*fun)(unsigned long))
{
      init_timer(tim);
      tim->expires = jiffies + 4*HZ;
      tim->data = (unsigned long)xsp;
      tim->function = fun;
      add_timer(tim);
}

void cancel_time_delay(struct timer_list*tim)
{
      del_timer(tim);
}

int wait_for_dispatch(struct Instance*xsp, struct ChannelData*xpd,
		      unsigned long mask)
{
      return atomic_sleep_on(xsp, mask, &xsp->dispatch_sync);
}

#ifndef DEVICE_MAX
# define DEVICE_MAX 2
#endif

# define FRAME_TABLE_MAGIC 0x2eaffaaf

static unsigned num_dev = 0;
const unsigned ucr_major = 38;

static struct Instance inst[DEVICE_MAX];

/*
 * The ucr_make_frame and ucr_free_frame functions perform the work of
 * allocating/releasing the frames. They check that the parameters are
 * reasonable, and do the allocation/release of memory. Allocate
 * PAGE_SIZE chunks because these are convenient for the mmap function.
 */

int ucr_make_frame(struct Instance*xsp, unsigned id, unsigned long size)
{
      unsigned idx;
      unsigned npages = (size + PAGE_SIZE - 1) / PAGE_SIZE;

	/* Allocate the page pointer block, and the pages
	   themselves. If there is a problem, jump to error exit
	   code. */

      size = npages * PAGE_SIZE;

      xsp->frame[id] = kmalloc(sizeof *xsp->frame[id] +
			       npages*sizeof(u32), GFP_KERNEL);
      if (xsp->frame[id] == 0)
	    return -ENOMEM;

      xsp->frame[id]->magic = FRAME_TABLE_MAGIC;
      xsp->frame[id]->self  = virt_to_bus(xsp->frame[id]);
      xsp->frame[id]->page_size  = PAGE_SIZE;
      xsp->frame[id]->page_count = (size + PAGE_SIZE - 1) / PAGE_SIZE;

      for (idx = 0 ;  idx < xsp->frame[id]->page_count ;  idx += 1) {

	    void*addr = (void*)get_free_page(GFP_KERNEL|GFP_DMA);
	    if (addr == 0)
		  goto no_mem;

	    xsp->frame[id]->page[idx] = virt_to_bus(addr);
	    mem_map_reserve(MAP_NR(addr));
      }


      return size;


no_mem: { /* Gack! Not enough memory, clean up what I did get and
	     return the correct error code. */
	    unsigned jdx;
	    for (jdx = 0 ;  jdx < idx ;  jdx += 1) {
		  unsigned long page;
		  page = (unsigned long)bus_to_virt(xsp->frame[id]->page[jdx]);
		  mem_map_unreserve(MAP_NR(page));
		  free_page(page);
	    }

	    kfree(xsp->frame[id]);
	    xsp->frame[id] = 0;
	    return -ENOMEM;
      }
}

int ucr_free_frame(struct Instance*xsp, unsigned id)
{
      unsigned idx;

      if (xsp->frame[id] == 0)
	    return -EIO;

      for (idx = 0 ;  idx < xsp->frame[id]->page_count ;  idx += 1) {
	    unsigned long page;
	    page = (unsigned long)bus_to_virt(xsp->frame[id]->page[idx]);
	    mem_map_unreserve(MAP_NR(page));
	    free_page(page);
      }

      kfree(xsp->frame[id]);
      xsp->frame[id] = 0;
      return 0;
}

static int xxopen(struct inode *inode, struct file *file)
{
      unsigned minor = MINOR(inode->i_rdev);
      int control_flag = (minor&0x80) != 0;
      struct Instance*xsp;
      int rc;

      minor &= 0x7f;

      if (minor >= num_dev) return -ENODEV;
      if (file->private_data) return -EIO;

      xsp = inst + minor;

      if (control_flag) {
	    rc =  ucrx_open(xsp);
	    if (rc != 0)
		  return rc;

      } else {

	    file->private_data = kmalloc(sizeof (struct ChannelData),
					 GFP_KERNEL);
	    if (file->private_data == 0)
		  return -ENOMEM;

	    rc = ucr_open(xsp, (struct ChannelData*)file->private_data);
	    if (rc != 0) {
		  kfree(file->private_data);
		  file->private_data = 0;
		  return rc;
	    }
      }

      MOD_INC_USE_COUNT;

      return 0;
}

static
#if LINUX_VERSION_CODE > 0x020100
int
#else
void
#endif
 xxrelease(struct inode *inode, struct file *file)
{
      unsigned minor = 0x7f & MINOR(inode->i_rdev);
      int control_flag = (0x80 & MINOR(inode->i_rdev)) != 0;
      struct Instance*xsp = inst + minor;
      struct ChannelData*xpd = (struct ChannelData*)file->private_data;

      if (control_flag) {
	    ucrx_release(xsp);
      } else {
	    ucr_release(xsp, xpd);
	    kfree(xpd);
      }

      MOD_DEC_USE_COUNT;
#if LINUX_VERSION_CODE > 0x020100
      return 0;
#endif
}

#if LINUX_VERSION_CODE >= 0x020142

static ssize_t xxread(struct file *file, char *bytes, size_t count, loff_t*off)
{
      long rc;
      unsigned minor = MINOR(file->f_dentry->d_inode->i_rdev) & 0x7f;
      int control_flag = (MINOR(file->f_dentry->d_inode->i_rdev) & 0x80) != 0;
      struct Instance*xsp = inst + minor;
      struct ChannelData*xpd = (struct ChannelData*)file->private_data;

      if (control_flag)
	    return -ENOSYS;

      rc = ucr_read(xsp, xpd, bytes, count, (file->f_flags&O_NONBLOCK)? 0 : 1);

      if (rc > 0)
	    *off += rc;

      return rc;
}

static ssize_t xxwrite(struct file*file, const char*bytes,
		       size_t count, loff_t*off)
{
      long rc;
      unsigned minor = MINOR(file->f_dentry->d_inode->i_rdev) & 0x7f;
      int control_flag = (MINOR(file->f_dentry->d_inode->i_rdev) & 0x80) != 0;
      struct Instance*xsp = inst + minor;
      struct ChannelData*xpd = (struct ChannelData*)file->private_data;

      if (control_flag)
	    return -ENOSYS;

      rc = ucr_write(xsp, xpd, bytes, count);
      if (rc > 0)
	    *off += rc;

      return rc;
}

#else

static long xxread(struct inode *inode, struct file *file,
		   char *bytes, unsigned long count)
{
      unsigned minor = MINOR(inode->i_rdev) & 0x7f;
      int control_flag = (MINOR(inode->i_rdev) & 0x80) != 0;
      struct Instance*xsp = inst + minor;
      struct ChannelData*xpd = (struct ChannelData*)file->private_data;

      if (control_flag)
	    return -ENOSYS;

      return ucr_read(xsp, xpd, bytes, count, (file->f_flags&O_NONBLOCK)? 0 : 1);
}

static long xxwrite(struct inode *inode, struct file *file,
		   const char*bytes, const unsigned long count)
{
      unsigned minor = MINOR(inode->i_rdev) & 0x7f;
      int control_flag = (MINOR(inode->i_rdev) & 0x80) != 0;
      struct Instance*xsp = inst + minor;
      struct ChannelData*xpd = (struct ChannelData*)file->private_data;

      if (control_flag)
	    return -ENOSYS;

      return ucr_write(xsp, xpd, bytes, count);
}
#endif

static int xxioctl(struct inode *inode, struct file *file,
		   unsigned int cmd, unsigned long arg)
{
      unsigned minor = MINOR(inode->i_rdev) & 0x7f;
      int control_flag = (MINOR(inode->i_rdev) & 0x80) != 0;
      struct Instance*xsp = inst + minor;
      struct ChannelData*xpd = (struct ChannelData*)file->private_data;

      if (control_flag)
	    return ucrx_ioctl(xsp, cmd, arg);
      else
	    return ucr_ioctl(xsp, xpd, cmd, arg);
}

static void xxirq(int irq, void *dev_id, struct pt_regs *pr)
{
      struct Instance*xsp = (struct Instance*)dev_id;
      ucr_irq(xsp);
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,1,0)
static int xxselect(struct inode*inode, struct file*file, int cmd,
		    select_table*wait)
{
      unsigned char minor = MINOR(inode->i_rdev) & 0x7f;
      struct Instance*xsp = inst + minor;
      struct ChannelData*xpd = (struct ChannelData*)file->private_data;
      int rc = 0;

      unsigned long flags;

      save_flags(flags);
      cli();

      switch (cmd) {

	  case SEL_IN:
	    if (xpd->table->first_in_idx != xpd->table->next_in_idx)
		  rc = 1;
	    else
		  select_wait(&xsp->dispatch_sync, wait);
	    break;

	  default: ;
      }

      restore_flags(flags);

      return rc;
}
#else
/*
 * Linux 2.1 introduces an improved form of select (called poll) that
 * expects the driver to return a bitmask of all the ready events on
 * the channel. The driver is also expected to register a wait event
 * in the poll table in case the thread must wait for something
 * useful.
 */
static unsigned int xxselect(struct file*file, poll_table*pt)
{
      unsigned minor = MINOR(file->f_dentry->d_inode->i_rdev) & 0x7f;
      struct Instance*xsp = inst + minor;
      struct ChannelData*xpd = (struct ChannelData*)file->private_data;

      poll_wait(file, &xsp->dispatch_sync, pt);
      if (! CHANNEL_IN_EMPTY(xpd))
	    return POLLIN | POLLRDNORM;

      return 0;
}
#endif

static void xxvmopen(struct vm_area_struct*vma)
{
	/* A new reference was spontaneously created. Count it. */
      MOD_INC_USE_COUNT;
}

static void xxvmclose(struct vm_area_struct*vma)
{
	/* A vma referencing the frame has disappeared. */
      MOD_DEC_USE_COUNT;
}

static void xxvmunmap(struct vm_area_struct *area, unsigned long off,
		      size_t size)
{
}

static unsigned long xxvmnopage(struct vm_area_struct*vma,
				unsigned long address,
				int write_access)
{
      unsigned long offset, page_nr;
      unsigned frame_nr;
#if LINUX_VERSION_CODE > 0x020100
      struct inode*inode = vma->vm_file->f_dentry->d_inode;
#else
      struct inode*inode = vma->vm_inode;
#endif
      unsigned minor = MINOR(inode->i_rdev);
      struct Instance*xsp = inst+minor;

      frame_nr = (vma->vm_offset >> 28) & 0x0f;
      offset = (address - vma->vm_start + vma->vm_offset) & 0x0fffffff;
      page_nr = offset / PAGE_SIZE;

      return (unsigned long)bus_to_virt(xsp->frame[frame_nr]->page[page_nr]);
}

static struct vm_operations_struct xxvm_ops = {
      xxvmopen,
      xxvmclose,
      xxvmunmap,
      0,
      0,
      0,
      xxvmnopage,
      0,
      0,
      0
};

/*
 * The mmap function places the 16 possible frames into a linear
 * address space with the high 4 bits used to select the frame. The
 * frame is created by the ioctl elsewhere.
 *
 * xxmmap will return an error if:
 *   The segment spans frames
 *   The segment oversteps the frame
 *
 */
#if LINUX_VERSION_CODE >= 0x020142
# define XXMMAP_PARMS struct file*file
#else
# define XXMMAP_PARMS struct inode*inode, struct file*file
#endif
static int xxmmap(XXMMAP_PARMS, struct vm_area_struct*vma)
{
      unsigned frame_nr;
      unsigned long offset, size, frame_size;
#if LINUX_VERSION_CODE >= 0x020142
      unsigned minor = MINOR(file->f_dentry->d_inode->i_rdev) & 0x7f;
#else
      unsigned minor = MINOR(inode->i_rdev) & 0x7f;
#endif
      struct Instance*xsp = inst+minor;

      vma->vm_flags |= VM_LOCKED;

      frame_nr = (vma->vm_offset >> 28) & 0x0f;
      offset = vma->vm_offset & 0x0fffffffU;
      size = vma->vm_end - vma->vm_start;

      if (frame_nr != ((vma->vm_offset+size) >> 28) & 0x0f) {
	    printk("ucr mmap: span frames %u and %u.\n", frame_nr,
		   ((vma->vm_offset+size) >> 28) & 0x0f);
	    return -EINVAL;
      }

      frame_size = xsp->frame[frame_nr]->page_size
	    * xsp->frame[frame_nr]->page_count;

      if ( (offset + size) > frame_size) {
	    printk("ucr mmap: size %lu too big for frame(%lu)\n",
		   (offset + size), frame_size);
	    return -EINVAL;
      }

      vma->vm_ops = &xxvm_ops;
#if LINUX_VERSION_CODE > 0x020100
      vma->vm_file = file;
#else
      vma->vm_inode = inode;
      inode->i_count++;
#endif
      MOD_INC_USE_COUNT;

      return 0;
}

static const struct file_operations ucr_ops = {
      0,
      xxread,
      xxwrite,
      0,
      xxselect,
      xxioctl,
      xxmmap,
      xxopen,
      xxrelease,
      0,
      0,
      0,
      0
};

int init_module(void)
{
      while (num_dev < DEVICE_MAX) {
	    unsigned char bus, device_fn;
	    unsigned addr, page;
	    unsigned long ptr;
	    struct Instance *xsp = inst + num_dev;

	      /* Test for the presence of yet another device. If it
		 does not exist, break out of the loop now. */

	    int rc = pcibios_find_device(VENDOR_ID, DEVICE_ID, num_dev,
					 &bus, &device_fn);
	    if (rc != PCIBIOS_SUCCESSFUL) break;

	    ucr_init_instance(xsp);
	    xsp->number = num_dev;
	    xsp->bus = bus;
	    xsp->dfn = device_fn;

	      /* Get the address of the DEVICE_NAME registers. Make sure the
		 memory is mapped into the kernel address space. By
		 default, it isn't necessarily. */

	    rc = pcibios_read_config_dword(bus, device_fn,
					   PCI_BASE_ADDRESS_0, &addr);
	    addr &= PCI_BASE_ADDRESS_MEM_MASK;

#if defined(__i386__)
	    page = addr & PAGE_MASK;
	    ptr = (unsigned long)ioremap(page, (addr - page) + DEVICE_SIZE);
	    xsp->dev = (ptr + (addr - page));
#else
	    xsp->dev = addr;
#endif

	      /* Get the interrupt number from the PCI config space,
		 and register an interrupt handler to deal with it. */
	    pcibios_read_config_byte(bus, device_fn,
				     PCI_INTERRUPT_LINE, &xsp->irq);
	    request_irq(xsp->irq, xxirq, SA_SHIRQ, DEVICE_NAME, xsp);

	    dev_init_hardware(xsp->dev);

	    printk(DEVICE_NAME "%u is a " DEVICE_LONG_NAME ".\n", num_dev);
	    num_dev += 1;
      }

      if (num_dev == 0) {
	    printk(DEVICE_NAME ": " DEVICE_LONG_NAME " not found.\n");
	    return -ENODEV;
      }

      if (debug_flag)
	    printk(DEVICE_NAME ": Debug_flag set to %u\n", debug_flag);

      if (register_chrdev(ucr_major, DEVICE_NAME, &ucr_ops)) {
	    printk(DEVICE_NAME ": unable to get major number %d\n", ucr_major);
	    return -EIO;
      }


      return 0;
}

void cleanup_module(void)
{
      unsigned idx;

      for (idx = 0 ;  idx < num_dev ;  idx += 1) {
	    struct Instance*xsp = inst + idx;
	    dev_clear_hardware(xsp->dev);
	    free_irq(xsp->irq, xsp);
	    ucr_clear_instance(xsp);
      }

      unregister_chrdev(ucr_major, DEVICE_NAME);
}

/*
 * $Log: linux.c,v $
 * Revision 1.4  1998/06/06 05:40:03  steve
 *  Linux/alpha 2.1.100 port.
 *
 * Revision 1.3  1998/05/29 20:57:12  steve
 *  build up frame table under NT.
 *
 * Revision 1.2  1998/05/28 22:53:02  steve
 *  NT port.
 *
 * Revision 1.1  1998/05/26 16:16:34  steve
 *  New channel protocol
 *
 */

