/*
 * Copyright (c) 1997 Picture Elements, Inc.
 *    Stephen Williams (steve@picturel.com)
 *
 *    This source code is free software; you can redistribute it
 *    and/or modify it in source code form under the terms of the GNU
 *    General Public License as published by the Free Software
 *    Foundation; either version 2 of the License, or (at your option)
 *    any later version. In order to redistribute the software in
 *    binary form, you will need a Picture Elements Binary Software
 *    License.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *  ---
 *    You should also have recieved a copy of the Picture Elements
 *    Binary Software License offer along with the source. This offer
 *    allows you to obtain the right to redistribute the software in
 *    binary (compiled) form. If you have not received it, contact
 *    Picture Elements, Inc., 777 Panoramic Way, Berkeley, CA 94704.
 */
#if !defined(WINNT)
#ident "$Id: winntdev.c,v 1.6 1998/07/11 20:46:39 steve Exp $"
#endif

/*
 * This file contains the interface functions for the ISEx
 * devices. The read/write/ioctl functions for actually accessing the
 * channels is implemented in this file.
 */
#include  "ucrif.h"
#include  "os.h"
#include  "t-config.h"
#include  "ucrpriv.h"

#define get_channel(irp) ((struct ChannelData*)(IoGetCurrentIrpStackLocation(irp)->FileObject->FsContext))


/*
 * This is the cancel function that handles enqueued read IRPs. All I
 * have to do here is remove from the pending_read_irps list and
 * cancel the IRP.
 */
static void dev_cancel_irp(DEVICE_OBJECT*dev, IRP*irp)
{
      struct Instance*xsp = (struct Instance*)dev->DeviceExtension;

      if (debug_flag & UCR_TRACE_CHAN) {
	    struct ChannelData*xpd = get_channel(irp);
	    printk(DEVICE_NAME "%u.%u (d): Cancel irp.\n",
		   xsp->number, xpd->channel);
      }

      RemoveEntryList(&irp->Tail.Overlay.ListEntry);

      IoReleaseCancelSpinLock(irp->CancelIrql);

      irp->IoStatus.Status = STATUS_CANCELLED;
      irp->IoStatus.Information = 0;
      IoCompleteRequest(irp, IO_NO_INCREMENT);
}

/*
 * Wait for dispatch will not actually wait, but will place the IRP on
 * the passed pend_list. It will mark that this has been done by
 * clearing the pend_list pointer.
 *
 * If there is no pend_list, then do block.
 */
int wait_for_dispatch(struct Instance*xsp, struct ChannelData*xpd,
		      unsigned long mask)
{
      if (debug_flag & UCR_TRACE_CHAN)
	    printk(DEVICE_NAME "%u.%u (d): Wait for dispatch. "
		   "pend_list=%p\n", xsp->number, xpd->channel,
		   xpd->pend_list);

      if (xpd->pend_list) {
	    KIRQL save_irql;
	    IoAcquireCancelSpinLock(&save_irql);
	    InsertTailList(xpd->pend_list, &xpd->irp->Tail.Overlay.ListEntry);
	    IoSetCancelRoutine(xpd->irp, dev_cancel_irp);
	    IoReleaseCancelSpinLock(save_irql);
	    xpd->pend_list = 0;
	    return -EINTR;
      }

      return atomic_sleep_on(xsp, mask, &xsp->dispatch_sync);
}


NTSTATUS dev_create(DEVICE_OBJECT*dev, IRP*irp)
{
      struct Instance*xsp = (struct Instance*)dev->DeviceExtension;
      struct ChannelData*xpd;
      int rc;

      xpd = ExAllocatePool(NonPagedPool, sizeof(struct ChannelData));
      if (xpd == 0) {
	    irp->IoStatus.Status = STATUS_NO_MEMORY;
	    irp->IoStatus.Information = 0;
	    IoCompleteRequest(irp, IO_NO_INCREMENT);
	    return STATUS_NO_MEMORY;
      }

      get_channel(irp) = xpd;

      xpd->block_flag = 0;
      xpd->pend_list = 0;
      rc = ucr_open(xsp, xpd);

      if (rc == 0) {
	    irp->IoStatus.Status = STATUS_SUCCESS;
	    irp->IoStatus.Information = 0;
	    IoCompleteRequest(irp, IO_NO_INCREMENT);
	    return STATUS_SUCCESS;
      }

      ExFreePool(xpd);
      get_channel(irp) = 0;

      switch (rc) {
	  default:
	    irp->IoStatus.Status = STATUS_UNSUCCESSFUL;
	    irp->IoStatus.Information = 0;
	    IoCompleteRequest(irp, IO_NO_INCREMENT);
      }

      return STATUS_UNSUCCESSFUL;
}

NTSTATUS dev_close(DEVICE_OBJECT*dev, IRP*irp)
{
      struct Instance*xsp = (struct Instance*)dev->DeviceExtension;
      struct ChannelData*xpd = get_channel(irp);

	/* Doesn't make a whole heck of a lot of sense to support
	   cancel of close, so invoke the blocking behavior. */
      xpd->irp = irp;
      xpd->pend_list = 0;
      ucr_release(xsp, xpd);

      ExFreePool(xpd);

      irp->IoStatus.Status = STATUS_SUCCESS;
      irp->IoStatus.Information = 0;
      IoCompleteRequest(irp, IO_NO_INCREMENT);
      return STATUS_SUCCESS;
}

NTSTATUS dev_read(DEVICE_OBJECT*dev, IRP*irp, BOOLEAN retry_flag)
{
      char*base;
      unsigned long cnt;
      KIRQL save_irql;
      int rc, flag;
      IO_STACK_LOCATION*stp = IoGetCurrentIrpStackLocation(irp);
      struct Instance*xsp = (struct Instance*)dev->DeviceExtension;
      struct ChannelData*xpd = get_channel(irp);

	/* Use the low bit of the block_flag flag to turn off blocking
	   for at least one byte. This works by making ucr_open return
	   0 instead of -EINTR if there is no data. */
      flag = (xpd->block_flag & 1) ? 0 : 1;

      if (! retry_flag)
	    stp->Parameters.Read.ByteOffset.LowPart = 0;


      base = MmGetSystemAddressForMdl(irp->MdlAddress);
      base += stp->Parameters.Read.ByteOffset.LowPart;
      cnt = stp->Parameters.Read.Length;
      cnt -= stp->Parameters.Read.ByteOffset.LowPart;

      xpd->irp = irp;
      xpd->pend_list = &xsp->pending_read_irps;
      rc = ucr_read(xsp, xpd, base, cnt, flag);

	/* Handle being placed on the pending list. */
      if (xpd->pend_list == 0) {
	    if (!retry_flag) IoMarkIrpPending(irp);
	    if (rc > 0) stp->Parameters.Read.ByteOffset.LowPart += rc;

	    if (debug_flag & UCR_TRACE_CHAN)
		  printk(DEVICE_NAME "%u.%u (d): read IRP marked pending\n",
			 xsp->number, xpd->channel);

	    return STATUS_PENDING;
      }

      xpd->pend_list = 0;

      if (rc >= 0) {
	    irp->IoStatus.Status = STATUS_SUCCESS;
	    irp->IoStatus.Information
		  = rc + stp->Parameters.Read.ByteOffset.LowPart;
	    IoCompleteRequest(irp, IO_NO_INCREMENT);
	    return STATUS_SUCCESS;
      }

      switch (rc) {

	  default:
	    irp->IoStatus.Status = STATUS_UNSUCCESSFUL;
	    break;
      }

      IoCompleteRequest(irp, IO_NO_INCREMENT);
      return irp->IoStatus.Status;
}

NTSTATUS dev_write(DEVICE_OBJECT*dev, IRP*irp, BOOLEAN retry_flag)
{
      const char*base;
      unsigned cnt;
      int rc;
      struct Instance*xsp = (struct Instance*)dev->DeviceExtension;
      struct ChannelData*xpd = get_channel(irp);
      IO_STACK_LOCATION*stp = IoGetCurrentIrpStackLocation(irp);

      if (! retry_flag)
	    stp->Parameters.Write.ByteOffset.LowPart = 0;

      xpd->irp = irp;
      xpd->pend_list = &xsp->pending_write_irps;

      base = MmGetSystemAddressForMdl(irp->MdlAddress);
      base += stp->Parameters.Write.ByteOffset.LowPart;
      cnt = stp->Parameters.Write.Length;
      cnt -= stp->Parameters.Write.ByteOffset.LowPart;

      rc = ucr_write(xsp, xpd, base, cnt);

	/* If I got put onto the pending list, the pend_list pointer
	   will have been cleared. */
      if (xpd->pend_list == 0) {
	    if (rc > 0) stp->Parameters.Write.ByteOffset.LowPart += rc;
	    if (!retry_flag) IoMarkIrpPending(irp);

	    if (debug_flag & UCR_TRACE_CHAN)
		  printk(DEVICE_NAME "%u.%u (d): write IRP marked pending\n",
			 xsp->number, xpd->channel);

	    return STATUS_PENDING;
      }

      xpd->pend_list = 0;
      if (rc >= 0) {
	    irp->IoStatus.Status = STATUS_SUCCESS;
	    irp->IoStatus.Information
		  = rc + stp->Parameters.Write.ByteOffset.LowPart;
	    IoCompleteRequest(irp, IO_NO_INCREMENT);
	    return STATUS_SUCCESS;
      }

      switch (rc) {

	  default:
	    irp->IoStatus.Status = STATUS_UNSUCCESSFUL;
	    break;
      }

      IoCompleteRequest(irp, IO_NO_INCREMENT);
      return irp->IoStatus.Status;
}

static NTSTATUS dev_mmap(struct Instance*xsp, IRP*irp)
{
      UNICODE_STRING phys_name;
      HANDLE phys_handle;
      OBJECT_ATTRIBUTES phys_attr;
      LARGE_INTEGER frame_base;
      NTSTATUS rc;

      IO_STACK_LOCATION*stp = IoGetCurrentIrpStackLocation(irp);
      struct UcrMmapInfo*mapinfo;

      mapinfo = (struct UcrMmapInfo*)irp->AssociatedIrp.SystemBuffer;
      if (mapinfo->frame_id >= 16) {
	    irp->IoStatus.Status = STATUS_UNSUCCESSFUL;
	    irp->IoStatus.Information = 0;
	    IoCompleteRequest(irp, IO_NO_INCREMENT);
	    return STATUS_UNSUCCESSFUL;
      }

      if (xsp->frame[mapinfo->frame_id] == 0) {
	    irp->IoStatus.Status = STATUS_NO_MEMORY;
	    irp->IoStatus.Information = 0;
	    IoCompleteRequest(irp, IO_NO_INCREMENT);
	    return STATUS_NO_MEMORY;
      }

      mapinfo->off  = 0;
      mapinfo->size = xsp->frame[mapinfo->frame_id]->page_count *
	    xsp->frame[mapinfo->frame_id]->page_size;

      RtlInitUnicodeString(&phys_name, L"\\Device\\PhysicalMemory");

      InitializeObjectAttributes(&phys_attr, &phys_name,
				 OBJ_CASE_INSENSITIVE, 0, 0);

      rc = ZwOpenSection(&phys_handle, SECTION_ALL_ACCESS, &phys_attr);
      if (rc != STATUS_SUCCESS) {
	    printk(DEVICE_NAME "%u: error opening physical memory.\n",
		   xsp->number);
	    irp->IoStatus.Status = rc;
	    irp->IoStatus.Information = 0;
	    IoCompleteRequest(irp, IO_NO_INCREMENT);
	    return rc;
      }

      mapinfo->base = 0;
      frame_base.HighPart = 0;
      frame_base.LowPart = xsp->frame[mapinfo->frame_id]->page[0];

      rc = ZwMapViewOfSection(phys_handle, (HANDLE)-1, &mapinfo->base, 0L,
			      mapinfo->size, &frame_base, &mapinfo->size,
			      ViewShare, 0, PAGE_READWRITE|PAGE_NOCACHE);
      if (rc != STATUS_SUCCESS) {
	    printk(DEVICE_NAME "%u: unable to map section. rc=%x, mapinfo->size=%x\n",
		    xsp->number, rc, mapinfo->size);
	    irp->IoStatus.Status = rc;
	    irp->IoStatus.Information = 0;
	    IoCompleteRequest(irp, IO_NO_INCREMENT);
	    return rc;
      }

      irp->IoStatus.Status = STATUS_SUCCESS;
      irp->IoStatus.Information = sizeof *mapinfo;
      IoCompleteRequest(irp, IO_NO_INCREMENT);

      return irp->IoStatus.Status;
}

static NTSTATUS dev_munmap(struct Instance*xsp, IRP*irp)
{
      struct UcrMmapInfo *mapinfo;

      mapinfo = (struct UcrMmapInfo*)irp->AssociatedIrp.SystemBuffer;
      ZwUnmapViewOfSection((HANDLE)-1, mapinfo->base);

      irp->IoStatus.Status = STATUS_SUCCESS;
      irp->IoStatus.Information = 0;
      IoCompleteRequest(irp, IO_NO_INCREMENT);

      return irp->IoStatus.Status;
}

static NTSTATUS dev_test(struct Instance*xsp, struct ChannelData*xpd, IRP*irp)
{
      unsigned idx;
      printk("XXXX First word of %u pages of frame 0...\n",
	     xsp->frame[0]->page_count);

      irp->IoStatus.Status = STATUS_SUCCESS;
      irp->IoStatus.Information = 0;
      IoCompleteRequest(irp, IO_NO_INCREMENT);

      return irp->IoStatus.Status;
}

/*
 * Process a DeviceIoControl call on the device. Get the input
 * argument and the command code, format them so that the generic code
 * can cope and let it fly.
 *
 * There are a few codes that are NT specific, so process them
 * directly.
 */
NTSTATUS dev_ioctl(DEVICE_OBJECT*dev, IRP*irp, BOOLEAN retry_flag)
{
      int rc;
      unsigned long cmd, arg;
      struct Instance*xsp = (struct Instance*)dev->DeviceExtension;
      struct ChannelData*xpd = get_channel(irp);
      IO_STACK_LOCATION*stp = IoGetCurrentIrpStackLocation(irp);

      cmd = stp->Parameters.DeviceIoControl.IoControlCode;

      switch (cmd) {

	  case UCR_MMAP_FRAME:
	    return dev_mmap(xsp, irp);
	  case UCR_MUNMAP_FRAME:
	    return dev_munmap(xsp, irp);

	  case UCR_NONBLOCK:
	    if ( (stp->Parameters.DeviceIoControl.InputBufferLength ==
		  sizeof(unsigned long))
		 && (*(unsigned long*)irp->AssociatedIrp.SystemBuffer) )
		  xpd->block_flag |= 1;
	    else
		  xpd->block_flag &= ~1;

	    rc = 0;
	    break;

	  case UCR_TEST:
	    return dev_test(xsp, xpd, irp);

	  case UCR_FLUSH:
	  case UCR_SYNC:
	      /* Certain of the UCR ioctls may be canceled. A little
		 bit of overhead is needed to support that behavior. */
	    xpd->irp = irp;
	    xpd->pend_list = &xsp->pending_ioctl_irps;
	    rc = ucr_ioctl(xsp, xpd, cmd, 0);
	    if (xpd->pend_list == 0) {
		  if (!retry_flag) IoMarkIrpPending(irp);
		  if (debug_flag & UCR_TRACE_CHAN)
			printk(DEVICE_NAME "%u.%u (d): ioctl(%x) IRP "
			       "marked pending\n", xsp->number,
			       xpd->channel, cmd);

		  return STATUS_PENDING;
	    }
	    xpd->pend_list = 0;
	    break;

	  default:
	      /* This is a standard UCR IOCTL, use the generic code to
		 process it. */

	    if (stp->Parameters.DeviceIoControl.InputBufferLength
		== sizeof arg)
		  arg = *(unsigned long*)irp->AssociatedIrp.SystemBuffer;
	    else
		  arg = 0;

	    xpd->pend_list = 0;
	    rc = ucr_ioctl(xsp, xpd, cmd, arg);
	    break;
      }

      if (rc < 0) {
	    irp->IoStatus.Status = STATUS_UNSUCCESSFUL;
	    irp->IoStatus.Information = 0;

      } else {
	    if (stp->Parameters.DeviceIoControl.OutputBufferLength >=
		sizeof(unsigned long)) {
		  *(unsigned long*)irp->AssociatedIrp.SystemBuffer = rc;
		  irp->IoStatus.Information = sizeof (unsigned long);
	    }
	    irp->IoStatus.Status = STATUS_SUCCESS;
      }

      IoCompleteRequest(irp, IO_NO_INCREMENT);
      return irp->IoStatus.Status;
}

/*
 * This is a DPC function scheduled by the interrupt handler. It
 * checks to see if there are any read IRPs stuck in the read queue
 * that can now go. It checks this by re-running them through the
 * dev_read function.
 */
static void recheck_pending_irps(KDPC*dpc, void*con, void*arg1, void*arg2)
{
      DEVICE_OBJECT*dev = (DEVICE_OBJECT*)con;
      struct Instance*xsp = (struct Instance*)dev->DeviceExtension;
      KIRQL save_irql;

      InitializeListHead(&xsp->tmp_queue);

      IoAcquireCancelSpinLock(&save_irql);
      while (! IsListEmpty(&xsp->pending_read_irps)) {
	    IRP*irp;
	    LIST_ENTRY*qe = RemoveHeadList(&xsp->pending_read_irps);

	    if (debug_flag & UCR_TRACE_CHAN) {
		  struct ChannelData*xpd = get_channel(irp);
		  printk("recheck read IRP\n");
	    }

	    irp = CONTAINING_RECORD(qe, IRP, Tail.Overlay.ListEntry);
	    IoSetCancelRoutine(irp, 0);
	    InsertTailList(&xsp->tmp_queue, qe);
      }
      IoReleaseCancelSpinLock(save_irql);

      while (! IsListEmpty(&xsp->tmp_queue)) {
	    IRP*irp;
	    LIST_ENTRY*qe = RemoveHeadList(&xsp->tmp_queue);
	    irp = CONTAINING_RECORD(qe, IRP, Tail.Overlay.ListEntry);
	    dev_read(dev, irp, TRUE);
      }

      IoAcquireCancelSpinLock(&save_irql);
      while (! IsListEmpty(&xsp->pending_write_irps)) {
	    IRP*irp;
	    LIST_ENTRY*qe = RemoveHeadList(&xsp->pending_write_irps);

	    if (debug_flag & UCR_TRACE_CHAN) {
		  struct ChannelData*xpd = get_channel(irp);
		  printk("recheck write IRP\n");
	    }

	    irp = CONTAINING_RECORD(qe, IRP, Tail.Overlay.ListEntry);
	    IoSetCancelRoutine(irp, 0);
	    InsertTailList(&xsp->tmp_queue, qe);
      }
      IoReleaseCancelSpinLock(save_irql);

      while (! IsListEmpty(&xsp->tmp_queue)) {
	    IRP*irp;
	    LIST_ENTRY*qe = RemoveHeadList(&xsp->tmp_queue);
	    irp = CONTAINING_RECORD(qe, IRP, Tail.Overlay.ListEntry);
	    dev_write(dev, irp, TRUE);
      }

      IoAcquireCancelSpinLock(&save_irql);
      while (! IsListEmpty(&xsp->pending_ioctl_irps)) {
	    IRP*irp;
	    LIST_ENTRY*qe = RemoveHeadList(&xsp->pending_ioctl_irps);

	    if (debug_flag & UCR_TRACE_CHAN) {
		  struct ChannelData*xpd = get_channel(irp);
		  printk("recheck ioctl IRP\n");
	    }

	    irp = CONTAINING_RECORD(qe, IRP, Tail.Overlay.ListEntry);
	    IoSetCancelRoutine(irp, 0);
	    InsertTailList(&xsp->tmp_queue, qe);
      }
      IoReleaseCancelSpinLock(save_irql);

      while (! IsListEmpty(&xsp->tmp_queue)) {
	    IRP*irp;
	    LIST_ENTRY*qe = RemoveHeadList(&xsp->tmp_queue);
	    irp = CONTAINING_RECORD(qe, IRP, Tail.Overlay.ListEntry);
	    dev_ioctl(dev, irp, TRUE);
      }


}

static BOOLEAN xxirq(PKINTERRUPT irq, void*dev_id)
{
      DEVICE_OBJECT*dev = (DEVICE_OBJECT*)dev_id;
      struct Instance*xsp = (struct Instance*)dev->DeviceExtension;
      ucr_irq(xsp);
      KeInsertQueueDpc(&xsp->pending_io_dpc, 0, 0);
      return TRUE;
}

#define dev_name_str L"\\Device\\ise"
#define dev_link_str L"\\DosDevices\\ISE"

void unload_device(DEVICE_OBJECT*dev)
{
      struct Instance*xsp = (struct Instance*)dev->DeviceExtension;

      UNICODE_STRING dev_link;
      wchar_t dev_link_buf[sizeof(dev_link_str) + 10];
      UNICODE_STRING number;
      wchar_t number_buf[10];

	/* Remove the symbolic link in \DosDevices */
      dev_link.Buffer = dev_link_buf;
      dev_link.MaximumLength = sizeof dev_link_buf;
      dev_link.Length = 0;
      RtlZeroMemory(dev_link_buf, sizeof dev_link_buf);
      RtlAppendUnicodeToString(&dev_link, dev_link_str);

      number.Buffer = number_buf;
      number.MaximumLength = sizeof(number_buf);
      number.Length = 0;
      RtlZeroMemory(number_buf, sizeof number_buf);
      RtlIntegerToUnicodeString(xsp->number, 10, &number);

      RtlAppendUnicodeStringToString(&dev_link, &number);
      IoDeleteSymbolicLink(&dev_link);

      IoDisconnectInterrupt(xsp->irq);
      if (xsp->dev_is_mmapped) MmUnmapIoSpace((void*)xsp->dev, DEVICE_SIZE);

      ucr_clear_instance(xsp);
}

/*
 * This method, given the driver object and an instance number,
 * creates a device object.
 */
NTSTATUS create_device(DRIVER_OBJECT*drv, unsigned bus_no,
		       unsigned slot, unsigned num)
{
      unsigned long rc;
      PCI_COMMON_CONFIG config;
      NTSTATUS status;
      struct Instance*xsp;

      DEVICE_OBJECT*dev;

      UNICODE_STRING dev_name, dev_link;
      wchar_t dev_name_buf[sizeof(dev_name_str) + 10];
      wchar_t dev_link_buf[sizeof(dev_link_str) + 10];
      UNICODE_STRING number;
      wchar_t number_buf[10];

	/* Now get ALL the PCI configuration space information for the
	   device. Use this to figure out stuff about the device that
	   is needed for the later I/O operations. */
      rc = HalGetBusData(PCIConfiguration, bus_no, slot, &config,
			 sizeof config);
      if (rc != sizeof config) return STATUS_UNSUCCESSFUL;


	/* Make the kernel device name (\Device\ise0) from the prefix
	   and the current instance number. Lots of UNICODE kruft. */
      dev_name.Buffer = dev_name_buf;
      dev_name.MaximumLength = sizeof(dev_name_buf);
      dev_name.Length = 0;
      RtlZeroMemory(dev_name_buf, sizeof dev_name_buf);
      RtlAppendUnicodeToString(&dev_name, dev_name_str);

      number.Buffer = number_buf;
      number.MaximumLength = sizeof(number_buf);
      number.Length = 0;
      RtlZeroMemory(number_buf, sizeof number_buf);
      RtlIntegerToUnicodeString(num, 10, &number);

      RtlAppendUnicodeStringToString(&dev_name, &number);


	/* Make the kernel device object and initialize the instance
	   structure with the basics. */
      status = IoCreateDevice(drv, sizeof(struct Instance), &dev_name,
			      FILE_DEVICE_ISE, 0, FALSE, &dev);
      if (! NT_SUCCESS(status)) return status;

      dev->Flags |= DO_DIRECT_IO;

      xsp = (struct Instance*)dev->DeviceExtension;
      ucr_init_instance(xsp);
      xsp->number = num;
      xsp->bus = bus_no;
      xsp->dfn = slot;

      InitializeListHead(&xsp->pending_read_irps);
      InitializeListHead(&xsp->pending_write_irps);
      InitializeListHead(&xsp->pending_ioctl_irps);
      KeInitializeDpc(&xsp->pending_io_dpc, recheck_pending_irps, dev);

      for (rc = 0 ;  rc < 16 ;  rc += 1)
	    xsp->fb[rc] = 0;

	/* Make the device accessible to WIN32 by creating a link in
	   \DosDevices to this driver. */
      dev_link.Buffer = dev_link_buf;
      dev_link.MaximumLength = sizeof dev_link_buf;
      dev_link.Length = 0;
      RtlZeroMemory(dev_link_buf, sizeof dev_link_buf);
      RtlAppendUnicodeToString(&dev_link, dev_link_str);

      RtlAppendUnicodeStringToString(&dev_link, &number);
      status = IoCreateSymbolicLink(&dev_link, &dev_name);
      if (! NT_SUCCESS(status)) {
	    printk("Failed to create device symbolic link.\n");
      }


	/* Get the address of the I/O registers for the device. The
	   target specific code uses this values to access its
	   registers with the readl and writel macros. */
      {
	    BOOLEAN rc;
	    unsigned tmp;
	    PHYSICAL_ADDRESS addr, scr;
	    addr.LowPart = config.u.type0.BaseAddresses[0] & ~0xf;
	    addr.HighPart = 0;

	      /* Ask NT for a bus to logical translation. NOTE that
		 NT/ix86 seems to sometimes get totally messed up and
		 will fail to translate an address, even though it is
		 a 1-to-1 map. BUG! Cope by guessing. */
	    rc = HalTranslateBusAddress(PCIBus, bus_no, addr, &tmp, &scr);
	    if (rc == FALSE) {
		  IO_ERROR_LOG_PACKET*el =
			IoAllocateErrorLogEntry(dev,
						sizeof(IO_ERROR_LOG_PACKET));
		  el->MajorFunctionCode = 0;
		  el->RetryCount = 0;
		  el->ErrorCode = UCR_HTBA_BUG;
		  el->UniqueErrorValue = 0;
		  el->FinalStatus = 0;
		  el->SequenceNumber = 0;
		  el->IoControlCode = 0;
		  el->DumpDataSize = sizeof(unsigned long);
		  el->DumpData[0] = addr.LowPart;
		  el->NumberOfStrings = 0;
		  IoWriteErrorLogEntry(el);
		  tmp = 0;
		  scr = addr;
	    } else if (debug_flag & (UCR_TRACE_CHAN|UCR_TRACE_PROTO)) {
		  printk("Translated %d/%p to logical %p:%p\n",
			 bus_no, addr.LowPart, scr.HighPart, scr.LowPart);
	    }
	    if (tmp == 0) {
		  xsp->dev = (__u32)MmMapIoSpace(scr, DEVICE_SIZE, FALSE);
		  xsp->dev_is_mmapped = 1;
	    } else {
		  xsp->dev = scr.LowPart;
		  xsp->dev_is_mmapped = 0;
	    }
      }

	/* Connect the interrupt to the interrupt handler. */
      {
	    unsigned long ivec;
	    unsigned char irq;
	    KAFFINITY affinity;

	    irq = config.u.type0.InterruptLine;
	    ivec = HalGetInterruptVector(PCIBus, bus_no, irq, irq,
					 &xsp->irql, &affinity);
	    IoConnectInterrupt(&xsp->irq, xxirq, dev, 0, ivec,
			       xsp->irql, xsp->irql, LevelSensitive,
			       TRUE, affinity, FALSE);
      }

	/* Finally, create a matchine control device. */
      create_devx(drv, xsp);

      return STATUS_SUCCESS;
}


/*
 * $Log: winntdev.c,v $
 * Revision 1.6  1998/07/11 20:46:39  steve
 *  Log HalTranslateBusAddress nonsense.
 *
 * Revision 1.5  1998/07/11 18:56:51  steve
 *  Frames for multiple devices.
 *
 * Revision 1.4  1998/07/10 22:54:23  steve
 *  Work around HalTranslateBusAddress bug.
 *
 * Revision 1.3  1998/05/30 18:59:23  steve
 *  Read retry gets offset right.
 *
 * Revision 1.2  1998/05/29 18:11:22  steve
 *  Proper cancel/abort behavior for NT.
 *
 * Revision 1.1  1998/05/28 22:53:02  steve
 *  NT port.
 *
 */

