/*
 * Copyright (c) 1998 Picture Elements, Inc.
 *    Stephen Williams (steve@picturel.com)
 *
 *    This source code is free software; you can redistribute it
 *    and/or modify it in source code form under the terms of the GNU
 *    General Public License as published by the Free Software
 *    Foundation; either version 2 of the License, or (at your option)
 *    any later version. In order to redistribute the software in
 *    binary form, you will need a Picture Elements Binary Software
 *    License.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *  ---
 *    You should also have recieved a copy of the Picture Elements
 *    Binary Software License offer along with the source. This offer
 *    allows you to obtain the right to redistribute the software in
 *    binary (compiled) form. If you have not received it, contact
 *    Picture Elements, Inc., 777 Panoramic Way, Berkeley, CA 94704.
 */
#if !defined(WINNT)
#ident "$Id: winntframe.c,v 1.4 1998/04/30 01:57:10 steve Exp $"
#endif

#include  "ucrif.h"
#include  "os.h"
#include  "t-config.h"
#include  "ucrpriv.h"

/*
 * These functions support the allocation of frame space under NT.
 *
 * Each frame is allocated as contiguous memory. The limitation here
 * is not the target, as it can scatter/gather, but the ability of NT
 * to map memory to the process. The ZwMapViewOfSection takes a memory
 * handle, and I can only seem to get that handle for physical
 * (and therefore contiguous) memory.
 */

static struct FramePool {
	  void*base;
	  unsigned long size;
} frame_pool[16];

int ucr_init_nt_frame_pool(unsigned long size)
{
        unsigned idx;
	PHYSICAL_ADDRESS limit;

	size = (size + PAGE_SIZE - 1) & ~(PAGE_SIZE-1);
	frame_pool[0].size = size;
	limit.LowPart = 0xffffffff;
	limit.HighPart = 0;
	frame_pool[0].base = MmAllocateContiguousMemory(size, limit);
	if (frame_pool[0].base)
	    return 1;
	else
	    return 0;

	for (idx = 1 ;  idx < 16 ;  idx += 1) {
		frame_pool[idx].size = 0;
		frame_pool[idx].base = 0;
	}
}

void ucr_deinit_nt_frame_pool()
{
      if (frame_pool[0].base)
	    MmFreeContiguousMemory(frame_pool[0].base);
}


int ucr_make_frame(struct Instance*xsp, unsigned id, unsigned long xsize)
{
      void*mem;
      unsigned idx;
      unsigned long fsize;

      if (xsp->frame[id].npages)
	    return xsp->frame[id].pagesize * xsp->frame[id].npages;

      if (frame_pool[id].base == 0)
	    goto no_mem;

      fsize = frame_pool[id].size;
      xsp->frame[id].pagesize = PAGE_SIZE;
      xsp->frame[id].npages = fsize / PAGE_SIZE;
      xsp->frame[id].pages = ExAllocatePool(NonPagedPool,
					    xsp->frame[id].npages *
					    sizeof(struct shared_buf));
      if (xsp->frame[id].pages == 0)
	    goto no_mem;

      mem = frame_pool[0].base;

      if (debug_flag & UCR_TRACE_FRAME)
	    printk("ise%u (d): frame segment at %x\n", xsp->number, mem);

      for (idx = 0 ;  idx < xsp->frame[id].npages ;  idx += 1) {
	    PHYSICAL_ADDRESS phys;

	    xsp->frame[id].pages[idx].host = (void*)
		  ((char*)mem + xsp->frame[id].pagesize * idx);
	    phys = MmGetPhysicalAddress(xsp->frame[id].pages[idx].host);
	    xsp->frame[id].pages[idx].bus = phys.LowPart;

	    if (debug_flag & UCR_TRACE_FRAME)
		  printk("ise%u (d): Frame page host=%x phys=%x\n",
			 xsp->number,
			 ((char*)mem + xsp->frame[id].pagesize * idx),
			 phys.LowPart);
      }

      return fsize;

 no_mem:
      xsp->frame[id].pagesize = 0;
      xsp->frame[id].npages = 0;
      xsp->frame[id].pages = 0;
      return -ENOMEM;
}

int ucr_free_frame(struct Instance*xsp, unsigned id)
{
      if (xsp->frame[id].npages == 0) return -EIO;

      ExFreePool(xsp->frame[id].pages);

      xsp->frame[id].pagesize = 0;
      xsp->frame[id].npages = 0;
      xsp->frame[id].pages = 0;

      return 0;
}


/*
 * $Log: winntframe.c,v $
 * Revision 1.4  1998/04/30 01:57:10  steve
 *  Canonical messages.
 *
 * Revision 1.3  1998/04/25 05:05:26  steve
 *  Preallocate frame space under NT.
 *
 * Revision 1.2  1998/03/23 23:40:22  steve
 *  frames work under NT.
 *
 * Revision 1.1  1998/03/23 00:07:38  steve
 *  initial frame support for winnt.
 *
 */

